#include <stdint.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <FL/Fl.H>
#include <Fl/Fl_Window.H>
#include <FL/Fl_File_Chooser.H>
#include "Fl/Fl_Export.H"
#include "Fl/fl_types.h"
#include "ui.h"
#include "avrsim.h"

progmem flash[0x10000];
uint8_t ram[0x10000];

UserInterface *ui;
int run=0;
int brkp=0;
fn_rec *fn_root;

long int f_cpu=1000000;
int max=0,pc=0,clockcycles=0;
int sourcefile,sourceline;
int level=0;
char fn_ioDefs[PATH_MAX],fn_hexFile[PATH_MAX],devicename[50];
char includePath[10*PATH_MAX];
uint16_t spl;
uint16_t sph;
uint16_t sreg;
uint16_t udr=0;
uint16_t sramstart=0x60,ramend=0x45f;
uint16_t flashend=0xff,ioend=0x3f;
char flags[]="ITHSVNZC";
char io_names[256][16];

#include "hardware.c"
#include "disass.c"
#include "simul.c"
#include "file_oper.c"

#ifndef WIN32
#include "images/avrsim.xpm"
static Fl_Pixmap p_sim(avr_sim_xpm);
static Fl_RGB_Image img_icon(&p_sim,0);
#endif

void actSource(int file,int line){
	int start,end,r,sX,sY;
	if(sourcefile!=file){
		sourcefile=file;
		ui->src_output->buffer()->loadfile(getSrcFileName(file));
		ui->src_win->label(getSrcFileName(file));
		sourceline=-1;
	}
	if(sourceline!=line){
		ui->src_output->buffer()->unselect();
		start=ui->src_output->buffer()->skip_lines(0,line-1);
		end=ui->src_output->buffer()->line_end(start);
		ui->src_output->buffer()->select(start,end);
		/* sicherstellen, dass die Zeile sichtbar ist */
		r=ui->src_output->position_to_xy(start,&sX,&sY);
		if((sY>ui->src_output->h()-30)||(!r)){
			ui->src_output->scroll(line-1,0);
		}
		sourceline=line;
	}
}

void Disass(){
	int op=0,op_old,l=0,c;
	char st[100],st1[100],brk;

	op=pc;
	ui->disass_output->buffer()->text("");
	ui->input_registers->buffer()->text("");
	while(l<60){
		op_old=op;
		disass(&op,st);
		if(op_old==brkp) brk='>'; else brk=' ';
		sprintf(st1,"%c%04X:",brk,op_old);
		c=0;
		while(op_old<op){
			sprintf(st1+strlen(st1)," %04X",flash[op_old].opcode);
			op_old++;
			c++;
		}
		while(c<3){
			strcat(st1,"     ");
			c++;
		}
		sprintf(st1+strlen(st1),"%s\n",st);
		ui->disass_output->buffer()->append(st1);
		l++;
	}
	l=ram[sreg];
	strcpy(st,"");
	for(c=0;c<8;c++){
		sprintf(st+strlen(st),"%c=%d ",flags[c],(l>>7)&1);
		l<<=1;
	}
	l=ram[spl];
	if(ram[sph]) l+=(ram[sph])<<8;
	sprintf(st+strlen(st),"  SP=%04x X=%04x Y=%04x Z=%04x",l,getPtr(X),getPtr(Y),getPtr(Z));
	strcat(st,"\n\n");
	ui->input_registers->buffer()->append(st);
	for(l=0;l<4;l++){
		st[0]=0;
		for(c=0;c<8;c++){
			sprintf(st+strlen(st),"r%02d=%02x  ",4*c+l,ram[4*c+l]);
		}
		strcat(st,"\n");
		ui->input_registers->buffer()->append(st);
	}
	ui->input_registers->buffer()->append("\n");
	op=32;	/* Beginn I/O-Bereich */
	l=0;
	while(op<96){
		sprintf(st,"%-6s=%02x  ",io_names[op],ram[op]);
		ui->input_registers->buffer()->append(st);
		op++;
		l++;
		if(l>5){
			ui->input_registers->buffer()->append("\n");
			l=0;
		}
	}
	l=0;
	op=sramstart;
	ui->ram_output->buffer()->text("");
	while(op<ramend+1){
		if(!(op&15)){
			sprintf(st,"%04X ",op);
			ui->ram_output->buffer()->append(st);
		}
		sprintf(st," %02X",ram[op]);
		ui->ram_output->buffer()->append(st);
		op++;
		if(!(op&15)){
			sprintf(st,"\n");
			ui->ram_output->buffer()->append(st);
		}
	}

	if(flash[pc].flags&PM_SOURCE){
		actSource(flash[pc].src_file,flash[pc].src_line);
	}
}

void actCyc(){
	char st[100];
	double d=0;
	sprintf(st,"%04X",brkp);
	ui->input_brkpt->value(st);
	sprintf(st,"%04X",pc);
	ui->input_pc->value(st);
	sprintf(st,"%d",clockcycles);
	ui->input_cycles->value(st);
	if(!f_cpu){
		ui->output_runtime->value("CPU Speed not defined");
	}
	if(clockcycles/f_cpu){
		d=(double)clockcycles/(double)f_cpu;
		sprintf(st,"%.2lf s",d);
		ui->output_runtime->value(st);
		return;
	}
	if(clockcycles*1000/f_cpu){
		d=(double)clockcycles/(double)f_cpu*1000;
		sprintf(st,"%.3lf ms",d);
		ui->output_runtime->value(st);
		return;
	}
	if(clockcycles*1000000/f_cpu){
		d=(double)clockcycles/(double)f_cpu*1000000;
		sprintf(st,"%.2lf µs",d);
		ui->output_runtime->value(st);
		return;
	}
	d=(double)clockcycles/(double)f_cpu*1000000000;
	sprintf(st,"%.1lf ns",d);
	ui->output_runtime->value(st);
}

void brkTest()
{
	if(pc==brkp){
		char st[100];
		run=0;
		sprintf(st,"Breakpoint 0x%04x reached",brkp);
		ui->statusline->value(st);
	}
}

void Reset(){
	run=0;
	pc=0;
	clockcycles=0;
	readCodeFile(fn_hexFile);
	actCyc();
	Disass();
}

void Stop(void){
	run=0;
	ui->statusline->value("Stopped");
}

void Step(){
	sim();
	actCyc();
	Disass();
}

void Run(){
	int c=0;
	run=1;
	ui->statusline->value("Running...");
	while(run){
		c++;
		if(c>10000){
			Fl::check();
			c=0;
		}
		sim();
		brkTest();
	}
	actCyc();
	Disass();
}

void StepOver(){
	int c=0,oldlevel=level;
	run=1;
	do{
		c++;
		if(c>10000){
			Fl::check();
			c=0;
		}
		sim();
		brkTest();
	} while((run)&&(oldlevel!=level));
	actCyc();
	Disass();
}

void StepOut(){
	int c=0,oldlevel=level;
	run=1;
	do{
		c++;
		if(c>10000){
			Fl::check();
			c=0;
		}
		sim();
		brkTest();
	} while((run)&&(oldlevel==level));
	actCyc();
	Disass();
}

void Autostep(){
	run=1;
	ui->statusline->value("Stepping...");
	while(run){
		Fl::check();
		sim();
		actCyc();
		Disass();
		brkTest();
	}
}

#include "ui.cxx"
#include "config.c"

int myHandler(int ev){
	int k,s;
	switch(ev){
		case FL_SHORTCUT:
		s=Fl::event_state();	/* shift keys */
		k=Fl::event_key();		/* key code */
		switch(k){
			case 0xffc2:	/* F5 */
			if(s&FL_SHIFT){
				Reset();
				return true;
			}
			if(s&FL_CTRL){
				Stop();
				return true;
			}
			if(s&FL_ALT){
				Autostep();
				return true;
			}
			Run();
			return true;
			case 0xffc3:	/* F6 */
			return true;
			case 0xffc4:	/* F7 */
			return true;
			case 0xffc5:	/* F8 */
			return true;
			case 0xffc6:	/* F9 */
			return true;
			case 0xffc7:	/* F10 */
			StepOver();
			return true;
			case 0xffc8:	/* F11 */
			if(s&FL_SHIFT){
				StepOut();
				return true;
			}
			Step();
			return true;
			case 0xffc9:	/* F12 */
			return true;
			default:
				//printf("Shortcut %x (%x)\n",k,s);
			break;
		}
		break;
	}
	return false;
}

int main(int argc, char **argv){
	FL_NORMAL_SIZE = 11;
	ui=new UserInterface();
	int i;
	char st[100];
	spl=0x5d;
	sph=0;
	sreg=0x5f;
	cleanFlash();
	for(i=0;i<256;i++) sprintf(io_names[i],"0x%02x",i);
	if(argc>1) readCodeFile(argv[1]);
	ui->disass_output->buffer(new Fl_Text_Buffer());
	ui->terminal_output->buffer(new Fl_Text_Buffer());
	ui->disass_output->buffer(new Fl_Text_Buffer());
	ui->ram_output->buffer(new Fl_Text_Buffer());
	ui->input_registers->buffer(new Fl_Text_Buffer());
	ui->src_output->buffer(new Fl_Text_Buffer());
	Fl::add_handler(myHandler);
	ui->btn_end->tooltip("Exit");
	ui->btn_run->tooltip("Run (F5)");
	ui->btn_brk->tooltip("Break(Ctrl-F5)");
	ui->btn_reset->tooltip("Reset (Shift-F5)");
	ui->btn_step->tooltip("Step Into (F11)");
	ui->btn_step_over->tooltip("Step Over (F10)");
	ui->btn_step_out->tooltip("Step Out (Shift-F11)");
	ui->btn_autostep->tooltip("Auto Step (Alt-F5)");
#ifndef WIN32
	ui->main_win->default_icon(&img_icon);
#endif
	readConfig();
	Disass();
	actCyc();
	ui->main_win->show();
	Fl::run();
	return 0;
}
